#include <stdio.h>
#include <assert.h>
#include <math.h>

#define MIN_N	1024		// 1KB
#define MAX_N	(8192*1024)	// 8MB
#define BASE	2

// sieve2
//
// return the number of prime numbers between 2 and n; v is an array
// with at least n elements
//
// This C function does the same thing as the optimized code in sieve.S,
// but it can be easily instrumented with calls to a memory hierarchy simulator
// to collect information on the simulated total number of extra cycles due to 
// cache and memory latency.  It counts the number of memory accesses, we can 
// change it to make a call to a function that returns the simulated latency of 
// a function that simulates an access to that array element's address in a 
// memory hierarchy 

unsigned long long int sieve2 (unsigned int n, unsigned char *v) {
	int i, j, lim, sum = 0;
	unsigned long long int accesses = 0;

	for (i=0; i<n; i++) {
		v[i] = 1;
		accesses++;
	}
	for (i=2; i<n/2; i++) {
		for (j=i*2; j<n; j+=i) {
			v[j] = 0;
			accesses++;
		}
	}
	for (i=2; i<n; i++) {
		sum += v[i];
		accesses++;
	}
	return accesses;
}

// num_insts
//
// this function returns the exact number of x86 instructions 
// executed by the sieve() function as it appears in sieve.S
// for an array of size n

int num_insts (int n) {
        int i, j, sum = 0;

	// compute number of instructions in inner loop

        for (i=2; i<n/2; i++) for (j=i*2; j<n; j+=i) sum += 7;

	// add up the rest of the instructions

        return 
                8 + 5 + 2 + 7   // instructions not in loops
                + n * 4         // first loop
                + 6 * (n/2-2)   // outer loop instructions
                + sum		// inner loop instructions
		+ 5 * (n-1);	// last loop
}

// a large array for the sieve and sieve2 functions

unsigned char bigarray[1<<24];

// testcache
//
// get the CPI for different sizes of the sieve array

void testcache (void) {
	// from sieve.S
	extern int sieve (int, unsigned char *);
	extern unsigned long long int read_tsc (void);
	int n, i = 0;

	// go through arrays of size MIN_N to MAX_N increasing by
	// a factor of BASE each time

	for (n=MIN_N; n<=MAX_N; n*=BASE) {
		// get the number of instructions executed by sieve() 
		// for an array of size n

		long long int m = num_insts (n);

		// total_insts will keep a running total of the number
		// of instructions executed; the goal is to let sieve run
		// for at least 1/2 billion instructions to get accurate
		// timing information

		long long int total_insts = 0;

		// t0 and t1 will measure the number of elapsed clock cycles
		// before and after running sieve

		unsigned long long int t0, t1;
		double cpi;

		// find out how many clock cycles have passed so far

		t0 = read_tsc ();

		// loop for at least 500,000,000 instructions

		for (;;) {

			// run the sieve function from sieve.S

			sieve (n, bigarray);

			// m more instructions have executed

			total_insts += m;

			// make sure we do this for at least 500,000,000 instructions
			// so we have accurate timing information

			if (total_insts >= 500000000) break;
		}
		
		// check the number of elapsed cycles again

		t1 = read_tsc ();

		// CPI is elapsed clock cycles divided by number of instructions

		cpi = (t1 - t0) / (double) total_insts;

		// print out the size of the array and the cpi

		printf ("n = %d, cpi = %f\n", n, cpi);
		fflush (stdout);
	}
}

// main program

int main () {
	testcache ();
	return 0;
}
